/* Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>, see file COPYING. */

/** \file afh.h Structures for audio format handling (para_server). */

/**
 * The tags used by all audio format handlers.
 *
 * Paraslash only uses the more common tags. These are recognized
 * for all supported audio formats.
 */
struct taginfo {
	/** TPE1 (id3v2) / ARTIST (vorbis) / ART (aac)/ author(spx) */
	char *artist;
	/** TIT2/TITLE/nam */
	char *title;
	/** TDRC/YEAR/day */
	char *year;
	/** TALB/ALBUM/alb */
	char *album;
	/** COMM/COMMENT/cmt */
	char *comment;
};

/** Audio format dependent information. */
struct afh_info {
	/** The number of chunks this audio file contains. */
	uint32_t chunks_total;
	/** The length of the audio file in seconds. */
	uint32_t seconds_total;
	/** Audio handler specific info about the file. */
	char *techinfo;
	/** Id3 tags, vorbis comments, aac tags. */
	struct taginfo tags;
	/**
	 * The table that specifies the offset of the individual pieces in
	 * the current audio file.
	 */
	uint32_t *chunk_table;
	/** Size of the largest chunk, introduced in v0.6.0. */
	uint32_t max_chunk_size;
	/** Period of time between sending data chunks. */
	struct timeval chunk_tv;
	/**
	 * The header is needed by senders in case a new client connects in the
	 * middle of the stream. The length of the header defaults to zero
	 * which means that this audio format does not need any special header
	 * treatment. The audio format handler does not need to set this to
	 * zero in this case.
	 */
	uint32_t header_len;
	/** The number of channels. */
	uint8_t channels;
	/** Frequency in Hz. */
	uint16_t frequency;
	/** Exact meaning depends on audio format. */
	uint16_t bitrate;
};

/** Data about the current audio file, passed from afs to server. */
struct audio_file_data {
	/** The open file descriptor to the current audio file. */
	int fd;
	/** Vss needs this for streaming. */
	struct afh_info afhi;
	/**
	 * Size of the largest chunk. Superseded by afhi->max_chunk_size. May
	 * be removed after v0.6.1.
	 */
	uint32_t max_chunk_size;
	/** Needed to get the audio file header. */
	uint8_t audio_format_id;
};

/**
 * Structure for audio format handling.
 *
 * There's one such struct for each supported audio format. Initially, only \a
 * name and \a init are defined. During the startup process, para_server calls
 * the \a init function of each audio format handler which is expected to fill
 * in the other part of this struct.
 */
struct audio_format_handler {
	/** Typical file endings for files that can be handled by this afh. */
	const char * const *suffixes;
	/**
	 * Check if this audio format handler can handle the file.
	 *
	 * This is a pointer to a function returning whether a given file is
	 * valid for this audio format. A negative return value indicates that
	 * this audio format handler is unable to decode the given file. On
	 * success, the function must return a positive value and fill in the
	 * given struct afh_info.
	 *
	 * \sa struct \ref afh_info.
	 */
	int (*get_file_info)(char *map, size_t numbytes, int fd,
		struct afh_info *afhi);
	/** Optional, used for header-rewriting. See \ref afh_get_header(). */
	void (*get_header)(void *map, size_t mapsize, char **buf, size_t *len);
	/**
	 * An audio format handler may signify support for dynamic chunks by
	 * defining ->get_chunk below. In this case the vss calls ->open() at
	 * BOS, ->get_chunk() for each chunk while streaming, and ->close() at
	 * EOS. The chunk table is not accessed at all.
	 *
	 * The function may return its (opaque) context through the last
	 * argument. The returned pointer is passed to subsequent calls to
	 * ->get_chunk() and ->close().
	 */
	int (*open)(const void *map, size_t mapsize, void **afh_context);
	/**
	 * Return a reference to one chunk. The returned pointer points to a
	 * portion of the memory mapped audio file. The caller must not call
	 * free() on it.
	 */
	int (*get_chunk)(long unsigned chunk_num, void *afh_context,
		const char **buf, size_t *len);
	/** Deallocate the resources occupied by ->open(). */
	void (*close)(void *afh_context);
	/**
	 * Write audio file with altered tags, optional.
	 *
	 * The output file descriptor has been opened by the caller and must not
	 * be closed in this function.
	 */
	int (*rewrite_tags)(const char *map, size_t mapsize, struct taginfo *tags,
		int output_fd, const char *filename);
};

int guess_audio_format(const char *name);
int compute_afhi(const char *path, char *data, size_t size,
	int fd, struct afh_info *afhi);
const char *audio_format_name(int);
__must_check int afh_get_chunk(long unsigned chunk_num, struct afh_info *afhi,
		uint8_t audio_format_id, const void *map, size_t mapsize,
		const char **buf, size_t *len, void **afh_context);
void afh_close(void *afh_context, uint8_t audio_format_id);
int32_t afh_get_start_chunk(int32_t approx_chunk_num,
		const struct afh_info *afhi, uint8_t audio_format_id);
void afh_get_header(struct afh_info *afhi, uint8_t audio_format_id,
		void *map, size_t mapsize, char **buf, size_t *len);
void afh_free_header(char *header_buf, uint8_t audio_format_id);
void clear_afhi(struct afh_info *afhi);
unsigned afh_get_afhi_txt(int audio_format_num, struct afh_info *afhi, char **result);
int afh_rewrite_tags(int audio_format_id, void *map, size_t mapsize,
		struct taginfo *tags, int output_fd, const char *filename);
void set_max_chunk_size(struct afh_info *afhi);
bool afh_supports_dynamic_chunks(int audio_format_id);
